@@ -1429,6 +1429,7 @@ gnome_rr_output_set_backlight (GnomeRROutput *output, gint value, GError **error
     return meta_dbus_display_config_call_change_backlight_sync (output->info->screen->priv->proxy,
 								output->info->serial,
 								output->id, value,
+								&output->backlight,
 								NULL, error);
 }
 
@@ -1475,8 +1476,7 @@ _gnome_rr_output_name_is_builtin_display (const char *name)
 	strstr (name, "LVDS") ||
 	strstr (name, "Lvds") ||
 	strstr (name, "LCD")  ||  /* ... but fglrx uses "LCD" in some versions.  Shoot me now, kthxbye. */
-	strstr (name, "eDP")  ||  /* eDP is for internal built-in panel connections */
-	strstr (name, "default")) /* Finally, NVidia and all others that don't bother to do RANDR properly */
+	strstr (name, "eDP"))     /* eDP is for internal built-in panel connections */
         return TRUE;
 
     return FALSE;
